# Copyright 2008 Neil Martinsen-Burrell

# This file is part of mcrepogen
#
# mcrepogen is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os, tempfile, sys

from bzrlib.workingtree import WorkingTree
from bzrlib.bzrdir import BzrDir
from bzrlib.revisionspec import RevisionSpec
from bzrlib.status import show_tree_status
from bzrlib.errors import InvalidRevisionSpec, BzrError
from bzrlib.plugin import load_plugins
from bzrlib.urlutils import unescape

from treebranch import Branch, DirectoryTree

class FastImportNotInstalled(BzrError):

    _fmt = ('To serialize trees, you must install bzr-fastimport.\n'
            'See https://launchpad.net/bzr-fastimport for more information.')


class BzrBranch(Branch):

    """Concrete implementation using Bazaar's branch structure"""

    def __init__(self, path=None, outf=sys.stderr):
        if path is None:
            location = tempfile.mkdtemp()
        else:
            location = path
        
        try:
            self._branch = BzrDir.open(location).open_branch()
        except:
            self._branch = BzrDir.create_branch_convenience(location)

        self.outf = outf

    def checkpoint(self, print_output=True):
        """Commit the current state of the working tree.
        
        :param print_output: bool
            If true, print the differences for each commit.  Otherwise, 
            dont print anything.  The stream used for printing is
            accessible as self.outf.
        """
        wt = self._branch.bzrdir.open_workingtree()
        if print_output:
            try:
                show_tree_status(wt, to_file=self.outf,
                    revision=[RevisionSpec.from_string('revno:-1')])
            except InvalidRevisionSpec:
                pass

        id = wt.commit('')
        if print_output:
            print >> self.outf, "Committed revision %s" % self._branch.last_revision_info()[0]
        return id

    def get_checkpoint(self, identifier):
        """Return a tree representing the state specified by the identifier."""
        pass

    def serialize(self, out_stream=sys.stdout):
        """Return a serialized version of the branch's history.
        
        :param out_stream: A stream for output.  It defaults to stdout.
        
        """
        load_plugins()
        try:
            from bzrlib.plugins.fastimport.bzr_exporter import BzrFastExporter
        except ImportError:
            raise FastImportNotInstalled
        exporter = BzrFastExporter(unescape(self._branch.base),
                                   None)
        exporter.outf = out_stream
        exporter.run()


class BzrTree(DirectoryTree):

    """Concrete implementation using Bazaar's working tree."""

    def __init__(self, path=None):
        """Create a bazaar tree.  Pass arguments on to WorkingTree"""
        if path is None: # create the tree in a temporary location
            location = tempfile.mkdtemp()
        else:
            location = path
        
        try:
            self._tree = WorkingTree.open(location)
        except:
            BzrDir.create_standalone_workingtree(location)
            self._tree = WorkingTree.open(location)

    def _absolute_path(self, relative_path):
        """Return the file system absolute path of a tree-relative path"""
        return self._tree.abspath(relative_path)

    def _safe_add(self, path):
        """Add a file or directory, returning the ID"""
        self._tree.lock_write()
        self._tree.add(path)
        self._tree.unlock()
        return self._tree.path2id(path)

    def id2path(self, id):
        """Return the full path for a specified entry"""
        return self._tree.id2path(id)

    def path2id(self, path):
        """Return the id for a specified full path"""
        return self._tree.path2id(path)

    def iter_files(self, directory=''):
        """Iterate over the files in a directory."""
        self._tree.lock_read()
        for child_entry in self._tree.walkdirs(directory).next()[1]:
            if child_entry[2] == 'file':
                yield child_entry[0]
        self._tree.unlock()

    def iter_subdirs(self, directory=''):
        """Iterate over the files in a directory."""
        self._tree.lock_read()
        for child_entry in self._tree.walkdirs(directory).next()[1]:
            if child_entry[2] == 'directory':
                yield child_entry[0]
        self._tree.unlock()

    def iter_all_files(self, directory=''):
        """Iterate recursively over all files in start"""
        self._tree.lock_read()
        for directory_entry in self._tree.walkdirs(directory):
            children = directory_entry[1]
            for child in children:
                if child[2] == 'file':
                    yield child[0]
        self._tree.unlock()

    def iter_all_subdirs(self, directory=''):
        """Iterate recursively over all directories under start"""
        self._tree.lock_read()
        for directory_entry in self._tree.walkdirs(directory):
            yield directory_entry[0][0]
        self._tree.unlock()

    def add_file(self, filepath, contents=''):
        """Add a file to the tree, returning its ID.

        To change the file after it is added, use get_file_by_id.
        """
        absolute_filepath = self._absolute_path(filepath)
        f = open(absolute_filepath,'w')
        f.write(contents)
        f.close()
        return self._safe_add(filepath)

    def add_directory(self, directory):
        """Add a directory to the tree, returning its ID."""
        absolute_directory = self._absolute_path(directory)
        os.mkdir(absolute_directory)
        return self._safe_add(directory)

    def _hard_remove(self, path):
        """Really remove a file from the tree and filesystem."""
        self._tree.lock_write()
        self._tree.remove(path, keep_files=False, force=True)
        self._tree.unlock()

    def remove_file(self, filepath):
        """Remove a file or directory from the tree."""
        self._hard_remove(filepath)

    def remove_directory(self, directory):
        """Remove a directory and its children from the tree."""
        children = list(self.iter_all_files(directory))
        children += list(self.iter_all_subdirs(directory))
        self._hard_remove(children)

    def move(self, path, destination):
        """Move the item at path to destination"""
        self._tree.lock_write()
        try:
            self._tree.rename_one(path, destination)
        finally:
            self._tree.unlock()

    def open_file(self, path, mode='r'):
        """Return a file handle to the given file.

        The file will be opened with the specified mode.
        """
        return open(self._absolute_path(path), mode)

    def open_file_by_id(self, id, mode='r'):
        """Return a file handle for the file specified by the id.

        The file may be opened in any of the allowed modes.
        """
        return open(self._tree.id2abspath(id), mode)
    

def create_branch_and_tree(path=None, *args, **kwargs):
    """Convenience function to create a Branch and Tree.

    Returns a tuple of (branch, tree).
    """
    if path is None:
        location = tempfile.mkdtemp()
    else:
        location = path
    BzrDir.create_standalone_workingtree(location)
    branch = BzrBranch(path=location, *args, **kwargs)
    tree = BzrTree(path=location)
    return branch, tree




#!/usr/bin/env python
# Copyright 2009 Neil Martinsen-Burrell

# This file is part of mcrepogen
#
# mcrepogen is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Create history of a random directory tree

MCREPOGEN creates random histories of directory trees using a Markov
Chain approach.
"""

import sys

import mcrepogen.transition as transition
from mcrepogen.version_info import version_info

def generate(output_file=None, revisions=100):
    b,t = transition.evolve(revisions)
    if output_file is None:
        out_stream = sys.stdout
    else:
        out_stream = file(output_file, 'w')
    b.serialize(out_stream)


if __name__ == "__main__":
    from optparse import OptionParser
    parser = OptionParser(usage="%prog [-r REVISIONS] [-o FILE]", 
                          version="%prog "+ "%d.%d" % version_info)
    parser.add_option('-o', '--output', dest='output', default=None,
                      help='Write output to FILE', metavar='FILE')
    parser.add_option('-r', '--revisions', dest='revisions', default=100,
                      type='int',
                      help='Number of revisions to create')

    (options, args) = parser.parse_args()
    generate(options.output, options.revisions)



# Copyright 2008 Neil Martinsen-Burrell

# This file is part of mcrepogen
#
# mcrepogen is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

class Branch(object):

    """Abstract Base Class for a branch (history of directory trees)"""

    def __init__(self, tree=None):
        """Create a branch whose current state is the given tree."""
        self.tree = tree

    def checkpoint(self):
        """Record the current state of the tree.

        Returns an identifier that can be used to retrieve the state of
        the tree at the time this method is called.
        """
        raise NotImplementedError

    def get_checkpoint(self, identifier):
        """Get a directory tree corresponding to the give identifier."""
        raise NotImplementedError

    def serialize(self):
        """A file-like object representing the history of the tree.
        
        Use the read() method to get the history.  The preferred format
        is the stream format used by git's fast-import.
        """
        raise NotImplementedError


class DirectoryTree(object):

    """Abstract Base Class for a directory tree."""

    def __init__(self):
        """Create a new DirectoryTree"""
        pass

    def id2path(self, id):
        """Return the full path for a specified entry"""
        raise NotImplementedError

    def path2id(self, path):
        """Return the id for a specified full path"""
        raise NotImplementedError

    def iter_files(self, directory=''):
        """Iterate over the paths of the files in a directory"""
        raise NotImplementedError

    def iter_subdirs(self, directory=''):
        """Iterate over the paths of the subdirectories in a directory"""
        raise NotImplementedError

    def iter_all_files(self, directory=''):
        """Iterate recursively over all files in directory"""
        raise NotImplementedError

    def iter_all_subdirs(self, directory=''):
        """Iterate recursively over all subdirectories of directory
        
        Note that this method yields an entry for the starting directory, in
        contrast to iter_subdirs.

        """
        raise NotImplementedError

    def add_file(self, filepath, contents=''):
        """Add a file to the tree, returning its ID.

        To change the file after it is added, use open_file_by_id.
        """
        raise NotImplementedError

    def add_directory(self, directory):
        """Add a directory to the tree, returning its ID"""
        raise NotImplementedError

    def remove_file(self, filepath):
        """Remove the file from the tree."""
        raise NotImplementedError

    def remove_directory(self, directory):
        """Remove a directory and its children from the tree."""
        raise NotImplementedError

    def move(self, from_path, to_path):
        """Move an item in the tree.

        Move the item that is currently at from_path to the new location
        to_path.  Thus renames and moving files around and moving
        directories around are all handled in the same way 

        """
        raise NotImplementedError

    def open_file(self, filepath, mode='r'):
        """Return a file object for the given file.

        The file will be opened with the specified mode.
        """
        raise NotImplementedError

    def open_file_by_id(self, id, mode='r'):
        """Return a file obect for the file specified by the id.

        The file may be opened in any of the allowed modes.
        """
        raise NotImplementedError





# Copyright 2008 Neil Martinsen-Burrell

# This file is part of mcrepogen
#
# mcrepogen is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys

from bzrtreebranch import create_branch_and_tree
from mutators import *
from editors import *

def evolve(iterations=100, mutators=None, editors=None):
    """Evolve a tree forward in time, snapshotting its progress.

    Create a tree and track its progress over time when changed by the
    specified mutators and editors.
    """

    if mutators is None:
        mutators = [FileCreator(),
                    DirectoryCreator(),
                    FileRemover(),
                    DirectoryRemover(),
                    FileMover(),
                    DirectoryMover(),
                   ]
    if editors is None:
        editors = [Patcher(),
                  ]

    transition = TransitionKernel(mutators, editors)
    branch, tree = create_branch_and_tree(outf=sys.stderr)

    ids = []
    for i in xrange(iterations):
        tree = transition.step(tree)
        new_id = branch.checkpoint()
        ids.append(new_id)
    return branch, tree


class TransitionKernel(object):

    """Transition object for a Markov Chain on directory trees."""

    def __init__(self, mutators, editors):
        """A TransitionKernel object with lists of mutators and editors.

        """
        self.mutators = mutators
        self.editors = editors

    def step(self, tree):
        """Take a forward time-step of the Markov Chain on the tree."""
        for mutator in self.mutators:
            tree = mutator.mutate(tree)
        for editor in self.editors:
            tree = editor.edit(tree)
        return tree



#!/usr/bin/env python
# Copyright 2009 Neil Martinsen-Burrell

# This file is part of mcrepogen
#
# mcrepogen is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from version_info import version_info
from distutils.core import setup

setup(name='MCREPOGEN',
      version='%d.%d' % version_info,
      description='Monte Carlo Repository Generator',
      long_description="""\
MCREPOGEN generates random version control histories. It uses a Markov Chain
model to generate successive states of a directory tree. You can generate as
many revisions as you want and the output is in the standard fast-import
format. This means that MCREPOGEN-generated histories can be imported into many
different version control systems to provide examples for testing.
""",
      author='Neil Martinsen-Burrell',
      author_email='nmb@wartburg.edu',
      url='http://launchpad.net/mcrepogen',
      download_url='http://launchpad.net/mcrepogen/+download',
      requires=['bzr', 'Numpy'],

      scripts=['gen_random_history.py'],
      packages=['mcrepogen'],
      package_dir={'mcrepogen' : '.'},
      package_data={'mcrepogen' : ['6of12.txt.gz']},

      classifiers=['Development Status :: 3 - Alpha',
                   'Environment :: Console',
                   'Intended Audience :: Developers',
                   'Intended Audience :: Science/Research',
                   'License :: OSI Approved :: GNU General Public License (GPL)',
                   'Programming Language :: Python :: 2',
                   'Topic :: Software Development :: Version Control',
                  ],
     )





# Copyright 2008 Neil Martinsen-Burrell

# This file is part of mcrepogen
#
# mcrepogen is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Objects that mutate the structure of the tree"""

from random import random, choice
import string, os, gzip
from string import whitespace

import numpy.random as mtrand

from bzrlib.errors import BzrMoveFailedError

word_list_location = os.path.join(os.path.dirname(__file__), '6of12.txt.gz')
WORD_LIST = [line.strip(whitespace+':&#<^=+') for line in
             gzip.GzipFile(word_list_location).readlines()]

def _random_filename():
    """Generate a random filename of specified length"""
    return choice(WORD_LIST)

def _random_content(length):
    """Returns a random line of content with at least length characters"""
    line = ''
    while len(line) < length:
        line += ' '+choice(WORD_LIST)
    return line


class Mutator(object):

    """This changes the contents of a tree in some probabilistic way."""

    def __init__(self, parameters):
        """Create an object with fixed probabilities.

        parameters should be a dictionary with the specified keys.  A
        value of None for any key will result in the default value being
        used.
        """
        if parameters is None:
            self.parameters = self.default_parameters
        else: # parameters should be a dictionary
            self.parameters = parameters
            for k,v in self.default_parameters.items():
                if self.parameters[k] is None:
                    self.parameters[k] = v

    def mutate(self, input_tree):
        """Return a changed version of the input_tree."""
        raise NotImplementedError


class FileCreator(Mutator):

    """A number of new files are created at random.

    The number of files created has a Poisson distribution (for the number of
    events occurring in a given time period, in this case between commits).
    The expected number of files created is lambda.

    Their initial content is a collection of a random number of lines
    with a random number of characters between 0 and max_line_length.
    The expected number of characters per line is max_line_length/2.
    
    The number of lines has a geometric distribution (how many lines do
    we have to write to get to the last one where the probability of
    each line being the last one is given).  The expected number of
    lines in a file is 1/p_last_line.
    
    """

    default_parameters = {'lambda': 1.9,
                          'max_line_length': 72,
                          'p_last_line': 0.025
                         }

    def __init__(self, parameters=None):
        """Create an object with fixed probabilities.

        parameters should be a dictionary with the following keys.  A
        value of None for any key will result in the default value being
        used.
    
        key                 default value
        ---                 -------------
        'lambda'                1.9
        'max_line_length'       72
        'p_last_line'           0.025
        """
        Mutator.__init__(self, parameters)

    def _random_content(self, num_lines):
        """Generate random content"""
        line_lengths = mtrand.randint(self.parameters['max_line_length'],
                                      size=num_lines)
        return '\n'.join(_random_content(length)
                         for length in line_lengths)

    def mutate(self, input_tree):
        """Add files with random initial content"""
        p = self.parameters['p_last_line']
        lam = self.parameters['lambda']
        all_dirs = list(input_tree.iter_all_subdirs())
        _geometric = mtrand.geometric
        for i in xrange(mtrand.poisson(lam)):
            dir = choice(all_dirs)
            if dir == '':
                full_path = _random_filename()
            else:
                full_path = dir + '/' + _random_filename()
            input_tree.add_file(full_path, 
                                self._random_content(_geometric(p)))
        return input_tree


class DirectoryCreator(Mutator):

    """Create subdirectories at random.

    The number of new directories created in has a Poisson distribution.

    """

    default_parameters = { 'lambda': 0.6 }

    def __init__(self, parameters=None):
        """Create an object with fixed probabilities.

        parameters should be a dictionary with the following keys.  A
        value of None for any key will result in the default value being
        used.
    
        key                 default value
        ---                 -------------
        'lambda'                0.6
        """
        Mutator.__init__(self, parameters)

    def mutate(self, input_tree):
        """Add random subdirectories"""
        subdirs = list(input_tree.iter_all_subdirs())
        num_to_add = mtrand.poisson(self.parameters['lambda'])
        for i in xrange(num_to_add):
            dir = choice(subdirs)
            if dir == '':
                new_path = _random_filename()
            else:
                new_path = dir + '/' + _random_filename()
            input_tree.add_directory(new_path)
        return input_tree


class FileRemover(Mutator):

    """Remove files at random.

    Each file is removed or not with a probability p, p << 1.  The
    expected number of files removed is N*p where N is the number of
    files in the tree.

    """

    default_parameters = { 'p': .01 }

    def __init__(self, parameters=None):
        """Create an object with fixed probabilities.

        parameters should be a dictionary with the following keys.  A
        value of None for any key will result in the default value being
        used.
    
        key                 default value
        ---                 -------------
        'p'                0.01
        """
        Mutator.__init__(self, parameters)

    def mutate(self, input_tree):
        """Remove files at random"""
        to_be_removed = []
        for filename in input_tree.iter_all_files():
            if mtrand.random() < self.parameters['p']:
                to_be_removed.append(filename)
        for filename in to_be_removed:
            input_tree.remove_file(filename)
        return input_tree

    
class FileMover(Mutator):

    """Move files to new names at random.
    
    Each file is moved to a randomly chosen directory with
    probability p, p << 1.  The expected number of files moved is N*p
    where N is the number of files in the tree.
    
    """

    default_parameters = {'p': 0.01}

    def __init__(self, parameters=None):
        """Create an object with fixed probabilities.

        parameters should be a dictionary with the following keys.  A
        value of None for any key will result in the default value being
        used.
    
        key                 default value
        ---                 -------------
        'p'                0.01
        """
        Mutator.__init__(self, parameters)

    def mutate(self, input_tree):
        """Move files around at random"""
        
        def _move(fromfile, to_dir):
            if os.path.dirname(fromfile) == to_dir:
                return
            to_path = os.path.join(to_dir,os.path.basename(fromfile))
            input_tree.move(fromfile, to_path)
        
        directories = list(input_tree.iter_all_subdirs())
        files = list(input_tree.iter_all_files())
        for f in files:
            if mtrand.random() < self.parameters['p']:
                _move(f, choice(directories))
        return input_tree


class DirectoryRemover(Mutator):

    """Remove entire directories at random.

    Each directory (and all of its contents) are removed with
    probability p << 1.  The expected number of directories removed is
    Nd*p where Nd is the number of directories in the tree.

    """

    default_parameters = {'p': .001}

    def __init__(self, parameters=None):
        """Create an object with fixed probabilities.

        parameters should be a dictionary with the following keys.  A
        value of None for any key will result in the default value being
        used.
    
        key                 default value
        ---                 -------------
        'p'                0.001
        """
        Mutator.__init__(self, parameters)

    def mutate(self, input_tree):
        """Remove directories at random"""
        to_remove = []
        for d in input_tree.iter_all_subdirs():
            if d == '':  #don't remove the root directory
                continue
            if mtrand.random() < self.parameters['p']:
                to_remove.append(d)
        for d in to_remove:
            input_tree.remove_directory(d)
        return input_tree


class DirectoryMover(Mutator):

    """Move directories to new names at random.
    
    Each directory is moved to a randomly chosen directory with
    probability p, p << 1.  The expected number of files moved is Nd*p
    where N is the number of files in the tree.
    
    """

    default_parameters = {'p': 0.001}

    def __init__(self, parameters=None):
        """Create an object with fixed probabilities.

        parameters should be a dictionary with the following keys.  A
        value of None for any key will result in the default value being
        used.
    
        key                 default value
        ---                 -------------
        'p'                0.001
        """
        Mutator.__init__(self, parameters)

    def mutate(self, input_tree):
        """Move directories around at random"""

        directories = list(input_tree.iter_all_subdirs())
        # Make two passes.  Choose the directories to move. Then choose
        # where to move them to for each directory.
        to_move = []
        for d in directories:
            if mtrand.random() < self.parameters['p']:
                to_move.append(d)
        while to_move:
            d = to_move.pop()
            new_home = choice(directories)
            try:
                input_tree.move(d, new_home)
            except BzrMoveFailedError:
                continue
            # rename everything involved with d
            new_path = os.path.join(new_home, os.basename(d))
            def _replacer(path):
                return string.replace(path, d, new_path)
            map(_replacer, directories)
            map(_replacer, to_move)
        return input_tree



# Copyright 2009 Neil Martinsen-Burrell

# This file is part of mcrepogen
#
# mcrepogen is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Objects that edit the content of files"""

import numpy.random as mtrand
import numpy as np
import random, string
from itertools import imap

from mutators import Mutator, _random_content

class Editor(Mutator):

    def edit(self, input_tree):
        raise NotImplmentedError

class Patcher(Editor):
    
    """Apply a randomly generated patch to randomly selected files.

    The number of files changed is a Poisson distribution with parameter lambda
    that gives the average number of files changed in a single revision.  

    The number of lines changed in the file is one plus a binomial with the
    number of lines is the file as n and a (small) probability p of each line
    being changed.  The changed lines are determined by a Bernoulli trial for
    each line and then split up into hunks based on their proximity.

    The number of lines replacing those in a hunk is Poisson with average
    equal to the number of lines in a hunk.

    The replacement lines have length uniformly distributed between 0 and
    max_line_length
    """

    default_parameters = {'lambda': 2,
                          'p_line_changed': .01,
                          'max_line_length': 72,
                         }

    def __init__(self, parameters=None):
        """Create an object with fixed probabilities.

        parameters should be a dictionary with the following keys.  A
        value of None for any key will result in the default value being
        used.
    
        key                 default value
        ---                 -------------
        'lambda'                2
        'p_line_changed'        .01
        'max_line_length'       72
        """
        Editor.__init__(self, parameters)


    def _patch_file(self, f):
        """Change the given file (opened for reading and writing)"""
        lines = f.readlines()
        length = len(lines)
        if length <= 1:
            changed_line_nums = [0]
        else:
            num_changed_lines = 1 + mtrand.binomial(length-1,
                self.parameters['p_line_changed'])
            changed_line_nums = random.sample(range(length), num_changed_lines)
        # make hunks
        changed_line_nums.sort()
        hunks = [[changed_line_nums[0]]]; hunk_num = 0
        for num in changed_line_nums[1:]:
            if num - hunks[hunk_num][-1] == 1:
                hunks[hunk_num].append(num)
            else: # start a new hunk
                hunks.append([num])
                hunk_num += 1
        # Change each hunk
        new_hunks = []
        for line_nums in hunks:
            new_hunk_length = mtrand.poisson(len(line_nums))
            line_lengths = mtrand.randint(self.parameters['max_line_length'],
                                          size=new_hunk_length)
            new_hunks.append([_random_content(length) + '\n'
                              for length in line_lengths])
        # Compose the new file
        f.seek(0); f.truncate()
        orig_file_pos = 0
        for i, line_nums in enumerate(hunks):
            f.writelines(lines[orig_file_pos:line_nums[0]])
            f.writelines(new_hunks[i])
            orig_file_pos = line_nums[-1] + 1
        f.writelines(lines[orig_file_pos:])

        f.close()


    def edit(self, input_tree):
        """Change lines in randomly selected files."""
        # select files that have changed
        n_changed = mtrand.poisson(self.parameters['lambda'], 1)[0]
        all_files = list(input_tree.iter_all_files())
        n_changed = min(n_changed, len(all_files))
        changed_files = random.sample(all_files, n_changed)
        for filename in changed_files:
            self._patch_file(input_tree.open_file(filename,'r+'))    
        return input_tree

